using System;
using System.Text;
using System.IO;

namespace JmBucknall.CsvParser {
  public class CsvParser {

    public void Parse(ICharTokenizer reader, ICsvConsumer consumer) {
      parseCsvFile(reader, consumer);
    }

    private void parseCsvFile(ICharTokenizer reader, ICsvConsumer consumer) {
      while (reader.Peek() != CsvConstants.EOF) {
        parseCsvRecord(reader, consumer);
      }
      consumer.SignalEndOfFile();
    }

    private void parseCsvRecord(ICharTokenizer reader, ICsvConsumer consumer) {
      parseCsvStringList(reader, consumer);
      char ch = reader.Read();
      if (ch == CsvConstants.EOF) {
        reader.Unread(ch);
        ch = '\n';
      }
      if (ch != '\n') {
        throw new CsvParserTooMuchDataException("End of record was expected but more data exists.");
      }
      consumer.SignalEndOfRecord();
    }
    
    private void parseCsvStringList(ICharTokenizer reader, ICsvConsumer consumer) {
      char ch;
      do {
        parseRawString(reader, consumer);
        ch = reader.Read();
      } while (ch == ',');
      reader.Unread(ch);
    }

    private bool isFieldTerminator(char c) {
      return ((c == ',') || (c == '\n') || (c == CsvConstants.EOF));
    }

    private bool isSpace(char c) {
      return ((c == ' ') | (c == '\t'));
    }

    private void parseOptionalSpaces(ICharTokenizer reader) {
      char ch;
      do {  
        ch = reader.Read();
      } while (isSpace(ch));
      reader.Unread(ch);
    }

    private void parseRawString(ICharTokenizer reader, ICsvConsumer consumer) {
      parseOptionalSpaces(reader);
      parseRawField(reader, consumer);
      if (!isFieldTerminator(reader.Peek()))
        parseOptionalSpaces(reader);
    }

    private void parseRawField(ICharTokenizer reader, ICsvConsumer consumer) {
      string fieldValue = String.Empty;

      char ch = reader.Peek();
      if (!isFieldTerminator(ch)) {
        if (ch == '"')
          fieldValue = parseQuotedField(reader);
        else 
          fieldValue = parseSimpleField(reader);
      }
      consumer.ConsumeField(fieldValue);
    }
 
    private string parseQuotedField(ICharTokenizer reader) {
      reader.Read(); // read and discard initial quote

      string field = parseEscapedField(reader);

      char ch = reader.Read();
      if (ch != '"') {
        reader.Unread(ch);
        throw new CsvParserNoTermQuoteException("Quoted field has no terminating double quote");
      }
      return field;
    }

    private string parseEscapedField(ICharTokenizer reader) {
      StringBuilder sb = new StringBuilder();

      parseSubField(reader, sb);
      char ch = reader.Read();
      while (processDoubleQuote(reader, ch)) {
        sb.Append('"');
        parseSubField(reader, sb);
        ch = reader.Read();
      }
      reader.Unread(ch);

      return sb.ToString();
    }

    private void parseSubField(ICharTokenizer reader, StringBuilder sb) {
      char ch = reader.Read();
      while ((ch != '"') && (ch != CsvConstants.EOF)) {
        sb.Append(ch);
        ch = reader.Read();
      }
      reader.Unread(ch);
    }

    private bool isBadSimpleFieldChar(char c) {
      return isSpace(c) || isFieldTerminator(c) || (c == '"');
    }

    private string parseSimpleField(ICharTokenizer reader) {

      char ch = reader.Read();
      if (isBadSimpleFieldChar(ch)) {
        reader.Unread(ch);
        return String.Empty;
      }

      StringBuilder sb = new StringBuilder();
      sb.Append(ch);
      ch = reader.Read();
      while (!isBadSimpleFieldChar(ch)) {
        sb.Append(ch);
        ch = reader.Read();
      }
      reader.Unread(ch);

      return sb.ToString();
    }

    private bool processDoubleQuote(ICharTokenizer reader, char ch) {
      if ((ch == '"') && (reader.Peek() == '"')) {
        reader.Read(); // discard second quote of double
        return true;
      }
      return false;
    }
  }
}