using System;
using JmBucknall.CsvParser;
using NUnit.Framework;

namespace JmBucknall.CsvParser.Test
{

  public class CsvConsumerTester : ICsvConsumer {

    private string[] expectedFields;
    private int fieldIndex;

    public CsvConsumerTester(params string[] expectedFields) {
      this.expectedFields = expectedFields;
    }

    public void ConsumeField(string s) {
      if (expectedFields[fieldIndex] != s) {
        throw new AssertionException(string.Format("field [{0}] expected, but [{1}] returned", expectedFields[fieldIndex], s));
      }
      fieldIndex++;
    }

    public void SignalEndOfRecord() {
      if (expectedFields[fieldIndex++] != "eor") {
        throw new AssertionException("End of record signalled but not expected");
      }
    }

    public void SignalEndOfFile() {
      if (expectedFields[fieldIndex++] != "eof") {
        throw new AssertionException("End of file signalled but not expected");
      }
    }
  }

  [TestFixture]
  public class TestCsvParser {

    [Test]
    public void TestEmptyInput() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("");
      CsvConsumerTester cct = new CsvConsumerTester("eof");

      cp.Parse(sct, cct);
    }
    [Test]
    public void TestSingleRecordOneField() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("a");
      CsvConsumerTester cct = new CsvConsumerTester("a", "eor", "eof");

      cp.Parse(sct, cct);
    }
    [Test]
    public void TestSingleRecordOneFieldWithEOR() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("a\n");
      CsvConsumerTester cct = new CsvConsumerTester("a", "eor", "eof");

      cp.Parse(sct, cct);
    }
    [Test]
    public void TestSingleRecordManyFields() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("a,b,c");
      CsvConsumerTester cct = new CsvConsumerTester("a", "b", "c", "eor", "eof");

      cp.Parse(sct, cct);
    }
    [Test]
    public void TestSingleRecordManyFieldsWithEOR() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("a,b,c\n");
      CsvConsumerTester cct = new CsvConsumerTester("a", "b", "c", "eor", "eof");

      cp.Parse(sct, cct);
    }
    [Test]
    public void TestManyRecordsManyFields() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("a,b\nc,d\ne,f");
      CsvConsumerTester cct = new CsvConsumerTester("a", "b", "eor", "c", "d", "eor", "e", "f", "eor", "eof");

      cp.Parse(sct, cct);
    }
    [Test]
    public void TestManyRecordsManyFieldsWithEOR() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("a,b\nc,d\ne,f\n");
      CsvConsumerTester cct = new CsvConsumerTester("a", "b", "eor", "c", "d", "eor", "e", "f", "eor", "eof");

      cp.Parse(sct, cct);
    }
    [Test]
    public void TestFieldsWithLeadingTrailingSpaces() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("     a,b     \nc     ,     d");
      CsvConsumerTester cct = new CsvConsumerTester("a", "b", "eor", "c", "d", "eor", "eof");

      cp.Parse(sct, cct);
    }
    [Test]
    public void TestQuotedFields() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("\"a b,c\",z\n\"s\nt\"\n\"a\"\"b\"\"\"");
      CsvConsumerTester cct = new CsvConsumerTester("a b,c", "z", "eor", "s\nt", "eor", "a\"b\"", "eor", "eof");

      cp.Parse(sct, cct);
    }
    [Test]
    public void TestEmptyQuotedField() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("\"\"");
      CsvConsumerTester cct = new CsvConsumerTester(String.Empty, "eor", "eof");

      cp.Parse(sct, cct);
    }
    [Test]
    [ExpectedException (typeof(CsvParserNoTermQuoteException))]
    public void TestQuotedFieldException() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("\"a b,c");
      CsvConsumerTester cct = new CsvConsumerTester("error: invalid token");

      cp.Parse(sct, cct);
    }
    [Test]
    [ExpectedException (typeof(CsvParserTooMuchDataException))]
    public void TestTooMuchDataException() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("\"a b,c\"z");
      CsvConsumerTester cct = new CsvConsumerTester("a b,c", "error: invalid token");

      cp.Parse(sct, cct);
    }
    [Test]
    [ExpectedException (typeof(CsvParserTooMuchDataException))]
    public void TestSimpleFieldWithQuote() {
      CsvParser cp = new CsvParser();
      StringCharTokenizer sct = new StringCharTokenizer("a\"z");
      CsvConsumerTester cct = new CsvConsumerTester("a", "error: invalid token");

      cp.Parse(sct, cct);
    }

  


  }
}
